#include-once
#include <GUIConstantsEx.au3>
#include <WindowsConstants.au3>
#include <GuiListView.au3>
#include <MsgBoxConstants.au3>
#include <StringConstants.au3>

#AutoIt3Wrapper_Au3Check_Parameters=-q -d -w 1 -w 2 -w 3 -w- 4 -w 5 -w 6 -w- 7

; #INDEX# =======================================================================================================================
; Title .........: Maps UDF
; AutoIt Version : 3.3.15.0 (Beta)
; Description ...: Experimental Maps UDF.
; Author(s) .....: Damon Harris (TheDcoder)
; Link ..........: https://www.autoitscript.com/forum/topic/178897-experimental-maps-udf/
; ===============================================================================================================================

; #CURRENT# =====================================================================================================================
; _Map_Append
; _Map_1DArrayToMap
; _Map_2DArrayToMap
; _Map_ConvertToArray
; _Map_Delete
; _Map_Display
; _Map_IniToMap
; _Map_MapToIni
; _Map_ReassignKey
; _Map_Remove
; ===============================================================================================================================

; #INTERNAL_USE_ONLY# ===========================================================================================================
; <None>
; ===============================================================================================================================

; #CONSTANTS# ===================================================================================================================
Global Const $MAP_CONVERT2DARRAY = 1
Global Const $MAP_CONVERT1DARRAY = 2
; ===============================================================================================================================

; #VARIABLES# ===================================================================================================================
; ...
; ===============================================================================================================================

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_Append
; Description ...: Append an element to a map.
; Syntax ........: _Map_Append(Byref $mMap, $vKey, $vContents)
; Parameters ....: $mMap                - [in/out] Map to proccess.
;                  $vKey                - The key to the element.
;                  $vContents           - Contents of the element.
; Return values .: Success: True
;                  Failure: False & @error set to non-zero
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......: Keys are case-sensitive.
; Related .......: MapAppend
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_Append(ByRef $mMap, $vKey, $vContents)
	If MapExists($mMap, $vKey) Then Return SetError(1, 0, False)
	$mMap[$vKey] = $vContents
	Return True
EndFunc   ;==>_Map_Append

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_1DArrayToMap
; Description ...: Converts a 1 dimensional array to a map.
; Syntax ........: _Map_1DArrayToMap(Byref $aArray, $vKeys[, $sDelimiter = ';'[, $iInitialElement = 0]])
; Parameters ....: $aArray              - [in/out] Array to convert.
;                  $vKeys               - A array containing the keys for each element OR
;                                         A string delimited with $sDelimiter containing the keys for each element.
;                  $sDelimiter          - [optional] Delimiter used to split the string (Ignored if an array). Default is ';'.
;                  $iInitialElement     - [optional] Element to take as a starting point. Default is 0.
; Return values .: Success: Converted Map
;                  Failure: False & @error set to:
;                           1 - If $vKeys is not an Array neither a String.
;                           2 - $aArray is not 1 dimensional.
;                           3 - If the no. of Elements in $aArray & $vKeys differ.
;                           4 - Conflicting keys were found, @extended contains the no. of conflicts.
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......:
; Related .......: _Map_2DArrayToMap
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_1DArrayToMap(ByRef $aArray, $vKeys, $sDelimiter = ';', $iInitialElement = 0)
	If IsString($vKeys) Then
		$vKeys = StringSplit($vKeys, $sDelimiter, $STR_NOCOUNT)
	ElseIf Not IsArray($vKeys) Then
		Return SetError(1, 0, False)
	EndIf
	If Not UBound($aArray, 0) = 1 Then Return SetError(2, 0, False)
	Local $iKeyCount = UBound($vKeys)
	Local $iElementCount = UBound($aArray)
	If Not $iKeyCount = ($iElementCount - $iInitialElement) Then Return SetError(3, 0, False)
	Local $mReturnMap[]
	Local $iErrorCount = 0
	For $iElement = $iInitialElement To ($iElementCount - 1) + $iInitialElement
		; $mReturnMap[$vKeys[$iElement]] = $aArray[$iElement]
		_Map_Append($mReturnMap, $vKeys[$iElement], $aArray[$iElement])
		$iErrorCount += @error
	Next
	If $iErrorCount > 0 Then Return SetError(4, $iErrorCount, False)
	Return $mReturnMap
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_2DArrayToMap
; Description ...: Converts a 2 dimensional array to a map.
; Syntax ........: _Map_2DArrayToMap(Byref $aArray[, $iKeyColumn = 0[, $iValueColumn = 1[, $iInitialRow = 0]]])
; Parameters ....: $aArray              - [in/out] Array to convert.
;                  $iKeyColumn          - [optional] The index of column containing the keys. Default is 0.
;                  $iValueColumn        - [optional] The index of column containing the values. Default is 1.
;                  $iInitialRow         - [optional] Row to take as a starting point. Default is 0.
; Return values .: Success: Converted Map.
;                  Failure: False & @error set to:
;                           1 - If $aArray is not 2 dimensional.
;                           2 - Conflicting keys were found, @extended contains the no. of conflicts.
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......:
; Related .......: _Map_1DArrayToMap
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_2DArrayToMap(ByRef $aArray, $iKeyColumn = 0, $iValueColumn = 1, $iInitialRow = 0)
	If Not UBound($aArray, 0) = 2 Then Return SetError(1, 0, False)
	Local $iRowCount = UBound($aArray)
	Local $mReturnMap[]
	Local $iErrorCount = 0
	For $iRow = $iInitialRow To ($iRowCount - 1) + $iInitialRow
		_Map_Append($mReturnMap, $aArray[$iRow][$iKeyColumn], $aArray[$iRow][$iValueColumn])
		$iErrorCount += @error
	Next
	If $iErrorCount > 0 Then Return SetError(2, $iErrorCount, False)
	Return $mReturnMap
EndFunc

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_ConvertToArray
; Description ...: Converts a Map to Array.
; Syntax ........: _Map_ConvertToArray(Byref $mMap[, $iArrayDimension = $MAP_CONVERT2DARRAY])
; Parameters ....: $mMap                - [in/out] Map to process.
;                  $iArrayDimension     - [optional] The dimension of the return array. Default is $MAP_CONVERT2DARRAY.
; Return values .: Success: A 1D Array containing all the contents of the map.
;                           A 2D Array containing key & value pair.
;                  Failure: False & @error set to non-zero
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......: 1. Only $MAP_CONVERT1DARRAY & $MAP_CONVERT2DARRAY are accepted in $iArrayDimension parameter, anything else sets
;                     @error to 1
;                  2. The 1st (0) element or the 2nd (1) column of 1st row contains the No. of keys/elements
; Related .......:
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_ConvertToArray(ByRef $mMap, $iArrayDimension = $MAP_CONVERT2DARRAY)
	Local $iKeyCount = UBound($mMap)
	Local $aKeys = MapKeys($mMap)
	Switch $iArrayDimension
		Case $MAP_CONVERT1DARRAY
			Local $aReturnArray[$iKeyCount + 1]
			$aReturnArray[0] = $iKeyCount
			For $iElement = 0 To $iKeyCount - 1
				$aReturnArray[$iElement + 1] = $mMap[$aKeys[$iElement]]
			Next
			Return $aReturnArray

		Case $MAP_CONVERT2DARRAY
			Local $aReturnArray[$iKeyCount + 1][2]
			$aReturnArray[0][0] = "No. of Rows"
			$aReturnArray[0][1] = $iKeyCount
			For $iRow = 1 To $iKeyCount
				$aReturnArray[$iRow][0] = $aKeys[$iRow - 1]
				$aReturnArray[$iRow][1] = $mMap[$aKeys[$iRow - 1]]
			Next
			Return $aReturnArray

		Case Else
			Return SetError(1, 0, False)
	EndSwitch
EndFunc   ;==>_Map_ConvertToArray

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_Delete
; Description ...: Deletes all the elements in a map.
; Syntax ........: _Map_Delete(Byref $mMap)
; Parameters ....: $mMap                - [in/out] Map to process.
; Return values .: Success: True & If $bHardWay is True Then @extended contains No. of elements in the map.
;                  Failure: False, @extended contains the Error Count & @error is set to non-zero.
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......: Only the elements are deleted, not the map itself! You can reuse the map if you wish.
; Related .......: MapRemove
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_Delete(ByRef $mMap, $bHardWay = False)
	If Not $bHardWay Then
		Local $mEmptyMap[]
		$mMap = $mEmptyMap
		Return True
	EndIf
	Local $iErrorCount = 0
	Local $aKeys = MapKeys($mMap)
	For $iKeyCount = 0 To UBound($aKeys) - 1
		MapRemove($mMap, $aKeys[$iKeyCount])
		If @error Then $iErrorCount += 1
	Next
	If $iErrorCount > 0 Then Return SetError(1, $iErrorCount, False)
	Return SetExtended($iKeyCount, True)
EndFunc   ;==>_Map_Delete

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_Display
; Description ...: Displays the Key & Value pair in a ListView.
; Syntax ........: _Map_Display(Byref $mMap[, $sTitle = "Map Display"])
; Parameters ....: $mMap                - [in/out] Map to display.
;                  $sTitle              - [optional] Title for the Display window. Default is "Map Display".
; Return values .: Success: True
;                  Failure: <N/A>
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......: This function is intended to be used for Debugging. Its still in development, so it does not provide flexible
;                  functionality like _ArrayDisplay
; Related .......: _ArrayDisplay
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_Display(ByRef $mMap, $sTitle = "Map Display")
	Local Const $esDataSepChar = Opt("GUIDataSeparatorChar")
	Local $hGUI = GUICreate($sTitle, 300, 300, -1, -1, BitOR($WS_MINIMIZEBOX, $WS_CAPTION, $WS_POPUP, $WS_SYSMENU, $WS_SIZEBOX, $WS_MAXIMIZEBOX))
	; Local $idTreeView = GUICtrlCreateTreeView(0, 0, 300, 264)
	Local $idListView = GUICtrlCreateListView("Key|Value", 0, 0, 300, 264)
	Local $idCopyButton = GUICtrlCreateButton("Copy Text", 3, 268, 146, 30)
	GUICtrlSetResizing(-1, $GUI_DOCKAUTO)
	Local $idExitButton = GUICtrlCreateButton("Exit Script", 149, 268, 148, 30)
	GUICtrlSetResizing(-1, $GUI_DOCKAUTO)

	Local $aKeys = MapKeys($mMap)
	Local $iKeyCount = UBound($mMap)
	Local $sContents = ""

	For $iKey = 0 To $iKeyCount - 1
		Switch VarGetType($mMap[$aKeys[$iKey]])
			Case "Map"
				$sContents = $aKeys[$iKey] & $esDataSepChar & '{Map}'

			Case "Array"
				$sContents = $aKeys[$iKey] & $esDataSepChar & '{Array}'

			Case Else
				$sContents = $aKeys[$iKey] & $esDataSepChar & $mMap[$aKeys[$iKey]]
		EndSwitch
		GUICtrlCreateListViewItem($sContents, $idListView)
	Next

	_GUICtrlListView_SetColumnWidth($idListView, 0, $LVSCW_AUTOSIZE)
	_GUICtrlListView_SetColumnWidth($idListView, 1, $LVSCW_AUTOSIZE)
	GUISetState()
	Local $nMsg = 0

	While 1
		$nMsg = GUIGetMsg()
		Switch $nMsg
			Case $GUI_EVENT_CLOSE
				GUIDelete($hGUI)
				Return True

			Case $idExitButton
				Exit

			Case $idCopyButton
				$sContents = GUICtrlRead(GUICtrlRead($idListView))
				ClipPut($sContents)
				MsgBox($MB_ICONINFORMATION, $sTitle, "The following has been copied to your clipboard: " & @CRLF & @CRLF & $sContents)
		EndSwitch
	WEnd
EndFunc   ;==>_Map_Display

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_IniToMap
; Description ...: Get a map representation of a Ini File.
; Syntax ........: _Map_IniToMap($sFile)
; Parameters ....: $sFile               - Path of the Ini File.
; Return values .: Success: Map representation of $sFile
;                  Failure: False & @error set to non-zero
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......: The map is in this format:
;                  $mIniFile["Section"]["Key"] = "Value"
; Related .......: _Map_MapToIni
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_IniToMap($sFile)
	If Not FileExists($sFile) Then
		Return SetError(1, 0, False)
	EndIf
	Local $mIniFile[], $mKeys[]
	Local $aSections = IniReadSectionNames($sFile)
	Local $aKeys
	For $iSection = 1 To $aSections[0]
		$aKeys = IniReadSection($sFile, $aSections[$iSection])
		For $iKey = 1 To $aKeys[0][0]
			$mKeys[$aKeys[$iKey][0]] = $aKeys[$iKey][1]
		Next
		$mIniFile[$aSections[$iSection]] = $mKeys
		_Map_Delete($mKeys)
	Next
	Return $mIniFile
EndFunc   ;==>_Map_IniToMap

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_MapToIni
; Description ...: Map to Ini File.
; Syntax ........: _Map_MapToIni(Byref $mIniFile, $sFile)
; Parameters ....: $mIniFile            - [in/out] Map representation of a INI File (See Remarks).
;                  $sFile               - Location of Ini File.
; Return values .: Success: True
;                  Failure: False & @error to non-zero
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......: 1. The map must be in this format: (same format as _Map_IniToMap)
;                     $mIniFile["Section"]["Key"] = "Value"
;                  2. The $sFile will be created if it does not already exist, albeit the root directory will not be created if
;                     it does not exist.
; Related .......: _Map_IniToMap
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_MapToIni(ByRef $mIniFile, $sFile)
	Local $aSections = MapKeys($mIniFile)
	Local $aKeys
	For $iSection = 0 To UBound($aSections) - 1
		$aKeys = MapKeys($mIniFile[$aSections[$iSection]])
		For $iKey = 0 To UBound($aKeys) - 1
			IniWrite($sFile, $aSections[$iSection], $aKeys[$iKey], $mIniFile[$aSections[$iSection]][$aKeys[$iKey]])
		Next
	Next
	Return True
EndFunc   ;==>_Map_MapToIni

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_ReasssignKey
; Description ...:
; Syntax ........: _Map_ReasssignKey(Byref $mMap, $vKey, $vNewKey)
; Parameters ....: $mMap                - [in/out] Map to process.
;                  $vKey                - Key to reassign.
;                  $vNewKey             - New key to assign.
; Return values .: Success: True
;                  Failure: False & @error set to non-zero
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......: Function will fail if $vKey does not exist or $vNewKey already exists
; Related .......:
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_ReassignKey(ByRef $mMap, $vKey, $vNewKey)
	If Not MapExists($mMap, $vKey) Or MapExists($mMap, $vNewKey) Then
		Return SetError(1, 0, False)
	EndIf
	Local $vContents = $mMap[$vKey]
	MapRemove($mMap, $vKey)
	$mMap[$vNewKey] = $vContents
	Return True
EndFunc   ;==>_Map_ReasssignKey

; #FUNCTION# ====================================================================================================================
; Name ..........: _Map_Remove
; Description ...: Removes an element from a map.
; Syntax ........: _Map_Remove(Byref $mMap, $vKey)
; Parameters ....: $mMap                - [in/out] Map to process.
;                  $vKey                - Key of the element to remove.
; Return values .: Success: True
;                  Failure: False & @error set to non-zero
; Author ........: Damon Harris (TheDcoder)
; Modified ......:
; Remarks .......: Keys are case-sensitive.
; Related .......: MapRemove
; Link ..........:
; Example .......: No
; ===============================================================================================================================
Func _Map_Remove(ByRef $mMap, $vKey)
	MapRemove($mMap, $vKey)
	If @error Then Return SetError(@error, 0, False)
	Return True
EndFunc   ;==>_Map_Remove